/* 
   hdcod v0.2 - a Hangul/English email folder decoder
  ====================================================

   This program 'hdcod' is personal purpose program, I programmed this just 
   for endless curiosity about 'Hangul Coding'. But I didn't understand
   the behind for 'Hangul Code'. I think this is a today's issue for our
   Korean Internet world must be solved emergently. I don't know much about
   computer and its coding method. Moreover I didn't understand Korean 
   language structure...
    
   Copyrights: 
       All the routine in this program follow the GNU Copying Policy. Please
       read the terms specified in GNU PUBLIC LICENSE. Temporarily I do not 
       include the terms here. Each routine's copyright is pertained to its
       original author described as follows:
          o ISO-2022-KR decoding - June-Yub Lee(1992.2.29)
          o Base64 decoding - Ienup Sung(1992.7.22)
          o Quoted Printable decoding - Park Myeong Seok(1995.4.12)
          o The others - Park Myeong Seok(1995.5.1,1996.6.8)

   Note: 
       A major feature of 'hdcod' is that no explicit code information is 
       required to decode a email or text folder. That is, this program does
       not require any character set informations. Strictly speaking, a mail
       implies one of coding informations in mail header or contents - this 
       is the basic idea for this program. This does not means 'hdcod' uses
       mail header informations such as 'Quoted Printable' for QP, contents
       character set 'iso-2022-kr' for ISO-2022-KR or 'base64' for 'Base64'.
       It looks sequential characters and judges what is the encoded method.
       Thus, it can correctly decode for one of the above three encoding 
       methods. If the given email or folder is already decoded or founded 
       unknown encoding method, 'hdcod' just dump it to an output. Just a 
       race condition of 'hdcod' is the one that the folder contains control
       characters. The line containing an unknown control character will be 
       skipped for safety but not completely - if possible avoid binary file
       as an input to 'hdcod'.
           You can modify the file IO to standard IO to use 'hdcod' as a 
       filter for some UMA(user mail agent) such as 'mail', 'elm' or 'pine'. 
       But this is not what I intended to. Current status of Korean MTA(mail
       transfer agent), i.e., 'Hangul sendmail' is in pending for argue about
       its usefulness or properness. If your environment is UNIX based, this
       program might be of no use. Already there exists useful de/encoding
       programs for general use of email decoding, e.g., 'hcode', 'metamail'. 
       But your system is not a UNIX clone such as M$ Window$ series, this 
       program is for you. Even you are using a UNIX clone, this program may 
       be useful if your system administrator is so ignorant in Korean mail 
       transfer.
           The name is similar to well-known 'hcode' because of its similar 
       usages, but you can find the differences. But, this decoding method 
       cannot be generalized. Only useful for encoded Korean and/or English 
       email. I intentionally use hexadecimal ASCII code to emphasize some
       characters are to be considered specially - you need ASCII table at 
       hand to read the source.

   Change log:
       1996. 6. 8. hdcod v0.2
           - NOTE: Changed copying policy(now follows GPL terms).
           - Rewrote hdcod v0.1.
           - Can use as a general Base64 and/or Quoted Printable and/or 
             ISO-2022-KR decoder.
           - Improved pattern checking methods: still buggie but almost safe. 
           - Improved mail header decoding and more stable QP decoding. 
           - Decode sequentially per line basis.
           - Bug for decoding a multiple folder was fixed.
       1995. 5. 1. hdcod v0.1
           - Rewrote mime2han. 
           - Mixed code decoder.
           - ISO-2022-KR & Base64 decoding were added. 
           - Decoding was performed sequentially per character basis.
       1995. 4.12. mime2han 
           - The original misleading name of hdcod was written. 
           - Supported only QP decoding.

   Usage: hdcod <encoded QP/B64/ISO Hangul/English email> <new KS output>

          Given a QP and/or B64 and/or ISO Korean/English mixed mail folder, 
          automatically converts it to KS code.
              QP: Quoted Printable, B64: Base64, ISO: ISO-2022-KR 
              KS: KSC-5601

   Bug report or any comments to: pms@romance.kaist.ac.kr
       - please include 'diff' result, if possible: do the 3 steps below.
         1. # hdcod input output
         2. # diff -u input output > error.bug
         3. And include the 'error.bug' file in your mail of bug report.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define BUF 1024
FILE *IN, *OUT;

static char *HelpText[] = {
  "\nhdcod v0.2: Decodes a QP/B64/ISO Hangul(English) folder to a KS folder.\n",
  "  (QP: Quoted Printable, B64: Base64, ISO: ISO-2022-KR, KS: KSC-5601)\n",
  "       - pms@romance.kaist.ac.kr\n\n",
  "Usage: hdcod <encoded QP/B64/ISO Hangul(English) email> <KS output>\n\n",
  ""
};

void HelpMsg(void)
{
    char **m = HelpText;

    while(**m) printf(*m++);
}

/*
**    Base64 decoding - Ienup Sung(1992.7.22)
*/
void DeB64(char *ob, const char *ib)
{
    register unsigned long bm=0;
    register short cnt=0;

    for(; *ib!=0x00; ib++) {
        bm = (*ib==0x2B)?  (bm << 6) | 0x3E:
             (*ib==0x2F)?  (bm << 6) | 0x3F:
             isdigit(*ib)? (bm << 6) | (0x34 + *ib - 0x30):
             isupper(*ib)? (bm << 6) | (*ib - 0x41):
             islower(*ib)? (bm << 6) | (0x1A + *ib - 0x61):
             (bm << 6) | 0;
        if(++cnt >= 4) {
            *ob++ = (char)((bm & 0x0FF0000) >> 16);
            *ob++ = (char)((bm & 0x000FF00) >> 8);
            *ob++ = (char) (bm & 0x00000FF);
            cnt = bm = 0;
        }
    }
    *ob = 0x00;
}

void DecodeB64(const char *s)
{
    char ts[BUF];

    DeB64(ts, s); fputs(ts, OUT);
}

/*
**    ISO-2022-KR decoding - June-Yub Lee(1992.2.29)
*/
void DecodeISO2022KR(const char *s)
{
    register i, kor, sl = strlen(s);    

    for(i=kor=0; i<sl; i++) {
        if(i<sl-3 & s[i]==0x1B & s[i+1]==0x24 & s[i+2]==0x29 & s[i+3]==0x43)
            i += 3;
        else if(s[i]==0x0E) kor = 1;
        else if(s[i]==0x0F) kor = 0;
        else if(kor) {
            if(s[i]==0x20|s[i]==0x09|s[i]==0x0A|s[i]==0x0D) fputc(s[i], OUT);
            else fputc(s[i] + 0x80, OUT);
        } 
        else fputc(s[i], OUT);
    }
}

int IsBin(char c)
{
    return 
        !(c==0x1B | c==0x0E | c==0x0F | c==0x09 | c==0x0A | c==0x0D)
        &((c>0x00 & c<0x20) | c==0x7F);
}
 
int IsB64(char c)
{
    return 
        (c==0x0A | c==0x0D)? 0x3001: (c==0x2B | c==0x2F | c==0x3D)? 0x4001: 
        isdigit(c)? 0x5001: isupper(c)? 0x6001: islower(c)? 0x7001: 0;
}

int IsWhite(char *s)
{
    while(*s++) {
        if(*s==0x09 | *s==0x0A | *s==0x0D | *s==0x20) continue;
        else return 0;
    }
    return 1;
}

/*
**    Quoted Printable decoding
*/
int DeQP(const char *s)
{
    register i, hc, hb;
    
    for(i=0; i<2; i++) {
        hb = (s[i] > 0x2F & s[i] < 0x3A)? s[i] - 0x30:
             (s[i] > 0x40 & s[i] < 0x47)? s[i] - 0x37:
             (s[i] > 0x60 & s[i] < 0x67)? s[i] - 0x57:
             -1;
        if(hb<0) return 0;
        hc = (i)? hc | hb: hb << 4; 
    }
    return hc;
}

/*
**  Quoted Printable decoding
*/
void DecodeQP(const char *s)
{
    char c, h[2], ns[BUF];
    register i, j, join, sl = strlen(s);

    for(i=join=0; i<sl; i++) {
        if(s[i] == 0x3D) {
            if(i==sl-2) { join = 1; break; }
            else if(i<sl-2) {
                for(j=0; j<2; j++) h[j] = s[i+1+j];
                if(!(c=DeQP(h))) fputc(0x3D, OUT);
                else { i+= 2; fputc(c, OUT); }
            }
        } else fputc(s[i], OUT);
    }
    if(join) {
        fgets(ns, BUF, IN);
        DecodeQP(ns);
    }
}

void Dump(const char *s)
{
    fputs(s, OUT);
}

int HeaderCode(char c)
{
    return (c==0x42 | c==0x62)? 4: (c==0x51 | c==0x71)? 3: 0;
}

void DeHeader(int code, const char *s)
{
    char ts[BUF];
    register i;

    if(code==3) DecodeQP(s);
    if(code==4) {
        DeB64(ts, s);
        for(i=0; i<strlen(ts); i++) fputc(ts[i], OUT);
    }
}

/*
**  Email Header Decoding
**
**      QP=3, B64=4
*/
void DecodeHeader(const char *s)
{
    char ts[BUF];
    register i, j, c, code, sl, fail;

    sl = strlen(s);
    for(i=code=0; i<sl; i++) {
        if(i<sl-1 & s[i]==0x3D & s[i+1]==0x3F) {
            fail = 0;
            if(i<sl-3 & s[i+3]==0x3F) {
                if(!code) code = HeaderCode(s[i+2]);
                if(code) {
                    for(j=0;j<sl-i-4 & s[i+4+j]!=0x3F;j++);
                    if(j>sl-i-4) fail = 1;
                    else c = i + 5 + j;
                } else fail = 1;
            }
            else {
                for(j=0; j<sl-i-2 & s[i+2+j]!=0x3F; j++);
                if(j<sl-i-3) {
                    if(!code) code = HeaderCode(s[i+3+j]); 
                    if(code && j<sl-i-4 & s[i+4+j]==0x3F) c = i + 5 + j;
                    else fail = 1;
                } else fail = 1;
            }
            if(!fail) {
                for(j=0;c<sl & s[c]!=0x3F; j++, c++) ts[j] = s[c];
                if(c<sl-1 & s[c+1]==0x3D) {
                    c++;
                    ts[j] = 0; DeHeader(code, ts);
                } else fail = 1;
            }
            if(fail) { 
                for(j=0; j<2; j++) fputc(s[i+j], OUT);
                i++;
            } else i = c;
        } else fputc(s[i], OUT); 
    }
}

/*
**  Check Code Pattern: depends on my knowledge of encoded email folder.
**                      - this means my method is not standard.
**             
**      DUMP=0, BIN=1, ISO2022KR=2, QP=3, B64=4, CODEDHEADER=5
*/
int CheckPattern(char *s)
{
    char c;
    register h, i, w, cqp, tqp, nqp, b64, eb64, nb64, cb64, pb64, sl, coh;
    
    for(i=0; i<(sl=strlen(s)) & ((c=s[i])==0x09 | c==0x20); i++); w = i;
    for(i=w, coh=tqp=cqp=b64=eb64=nb64=cb64=pb64=0; i<sl; i++, pb64=cb64) {
        if(IsBin(s[i])) return 1;
        if(i<sl-3 & s[i]==0x1B & s[i+1]==0x24 & s[i+2]==0x29 & s[i+3]==0x43)
            return 2;
        if(s[i]==0x0E | s[i]==0x0F) return 2;
        if(s[i]==0x3D) cqp++;
        if(i<sl-5 & s[i]==0x3D) {
            if(!tqp) tqp = (s[i+1]==0x33 & s[i+2]==0x44);
            nqp = isxdigit(s[i+1]) + isxdigit(s[i+2]) 
                                   + isxdigit(s[i+4]) + isxdigit(s[i+5]);
            if(nqp==4 & s[i+3]==0x3D) return 3;
        }
        if(s[i]==0x3D & s[i+1]==0x3F) coh = 1;
        for(h=i+2; coh && h<sl; h++) {
            if(s[h]==0x3F) coh++;
            if(coh > 2) return 5;
        }
        nb64 += (cb64 = IsB64(s[i])) & 1;
        if(i>w+1 & !b64 & (pb64!=0x5001 & cb64!=pb64) & cb64!=0x3001) b64 = 1;
        if(b64) eb64 += (s[i]==0x2F & i<20); 
    }
    if(eb64==1 & sl>10) eb64 = 0; 
    if(!eb64 & b64 & nb64==sl-w & sl>w+1 & w<2) return 4;
    if((s[sl-2]==0x3D & cqp==1) | tqp) return 3;
    return 0;
}

void Decode(int p, char *s)
{
    switch(p) {
        case 0:
            Dump(s); break;
        case 1:
            break;
        case 2:
            DecodeISO2022KR(s); break;
        case 3:
            DecodeQP(s); break;
        case 4:
            DecodeB64(s); break;
        case 5:
            DecodeHeader(s); break;
        default:
            break;
    }
}

int main(char ac, char **av)
{
    char c;
    static char s[BUF], ts[BUF];
    register wht, pat;

    if(ac != 3) { HelpMsg(); return 0; }
    if(!(IN=fopen(av[1], "r")))  { perror(av[1]); return 0; }
    if(!(OUT=fopen(av[2], "w"))) { perror(av[2]); return 0; }

    for(wht=pat=0; fgets(s, BUF, IN)!=NULL; *s=0) {
        if(wht & pat==4) { wht = 0; continue; }
        if(wht) { Dump(ts); ts[0] = 0; }  
        if((wht=IsWhite(s))>0) strcpy(ts, s);
        pat = CheckPattern(s);
        Decode(pat, s);
    }

    fclose(IN); fclose(OUT);
    return 0;
}
